import re
import sys
from .common import VlabException
from .vlab import Vlab
from threading import Event


def init_vlab(args):
    registers_trace = False
    functions_trace = False
    interrupts_trace = False
    if args.traces_list:
        for trace in args.traces_list.split(','):
            if trace == 'registers' or trace == 'regs':
                registers_trace = True
            elif trace == 'functions':
                functions_trace = True
            elif trace == 'interrupts':
                interrupts_trace = True
            else:
                raise VlabException(
                    'jumper run: error: Invalid trace type {}. Valid traces are regs, functions, interrupts'.format(
                        trace), 1)

    vlab = Vlab(
        working_directory=args.working_directory,
        sudo_mode=args.sudo_mode,
        gdb_mode=args.gdb_mode,
        registers_trace=registers_trace,
        functions_trace=functions_trace,
        interrupts_trace=interrupts_trace,
        trace_output_file=args.trace_output_file,
        print_uart=args.print_uart,
        uart_output_file=args.uart_output_file,
        uart_device=args.uart_device,
        platform=args.platform,
        token=args.token,
        debug_peripheral=args.debug_peripheral,
        gpio=args.gpio
    )

    vlab.load(args.fw)

    return vlab


def stop_after_parse(stop_after):
    def raise_invalid_time():
        print('Invalid time format for --stop-after: {}'.format(stop_after))
        return None

    match = re.match('^\d+',stop_after)
    if not match:
        return raise_invalid_time()
    ns = int(match.group(0))

    if re.match('^\d+$', stop_after) or re.match('^\d+ms$', stop_after):
        ns = ns * 1000 * 1000
    elif re.match('^\d+s$', stop_after):
        ns = ns * 1000 * 1000 * 1000
    elif re.match('^\d+us$', stop_after):
        ns = ns * 1000
    elif re.match('^\d+ns$', stop_after):
        pass
    else:
        return raise_invalid_time()

    return ns


class VlabRunner:
    def __init__(self):
        self.stop = Event()
        self.stop.clear()
        pass

    def _exit_callback(self, error_code, error_msg):
        # error_code & error_msg not it use here, but part of callback interface

        self.stop_run()

    def start(self, args):
        try:
            def bkpt_callback(code):
                print('Firmware reached a BKPT instruction with code {}'.format(code))
                self.stop.set()

            vlab = init_vlab(args)
            vlab.on_bkpt(bkpt_callback)
            vlab.register_on_exit(self._exit_callback)
            ns = None

            if args.stop_after:
                vlab._send_stop_after_event()

                ns = stop_after_parse(args.stop_after)
                if ns is None:
                    return 1

            vlab.start(0)

            if not args.print_uart and not args.traces_list:
                print(
                    '\nVirtual device is running without UART/Trace prints (use -u and/or -t to get your firmware '
                    'execution status)\n')
            else:
                print('\nVirtual device is running.\n')
                sys.stdout.flush()


            if ns:
                vlab.SUDO.set_timer(ns, self.stop_run)
            vlab.resume()

            self.wait_to_finish()

            vlab.stop()
            return vlab.get_return_code()

        except VlabException as e:
            print('jumper run: error: {}'.format(e.message))
            try:
                if vlab:
                    vlab.stop()
            finally:
                return e.exit_code

    def stop_wait(self):
        while not self.stop.wait(timeout=0.1):
            pass

    def stop_run(self):
        self.stop.set()

    def wait_to_finish(self):
        self.stop.wait()


"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from __future__ import print_function

import os
import sys
from time import sleep
import calendar
import time
import requests
import logging
import Queue
import threading
from __version__ import __version__ as jumper_current_version

API_URL = 'https://vlab.jumper.io/api/v2'
if 'JUMPER_STAGING' in os.environ:
    API_URL = 'https://us-central1-jemu-web-app-staging.cloudfunctions.net/api/v2'
if 'JUMPER_STAGING_INBAR' in os.environ:
    API_URL = 'https://us-central1-jemu-web-app-inbar.cloudfunctions.net/api/v2'


class WebException(Exception):
    pass


class AuthorizationError(WebException):
    def __init__(self, message):
        super(WebException, self).__init__(message)
        self.exit_code = 4
        self.message = message


class UnInitializedError(WebException):
    def __init__(self):
        print("Failed to get user id. Please reach out to support@jumper.io for help")
        super(WebException, self).__init__("Failed to get user id. Please reach out to support@jumper.io for help")
        self.exit_code = 6
        self.message = "Failed to get user id. Please reach out to support@jumper.io for help"


class EmulatorGenerationError(WebException):
    def __init__(self, message):
        super(WebException, self).__init__(message)
        self.exit_code = 5
        self.message = message


class JemuWebApi(object):
    def __init__(self, jumper_token=None, api_url=API_URL, local_jemu=None):
        self._api_url = api_url
        self._token = jumper_token
        self._headers = {'Authorization': 'Bearer ' + self._token}
        self._user_uid = None
        self._local_jemu = local_jemu

        logging.getLogger("requests").setLevel(logging.WARNING)
        res = requests.get(self._api_url + '/hello', headers=self._headers)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == requests.codes['unauthorized'] or res.status_code == requests.codes['forbidden']:
                print("Error: Authorization failed. Check the token in your config.json file")
                raise AuthorizationError("Error: Authorization failed. Check the token in the config.json file.")
            else:
                raise e

        self._user_uid = res.json()['userUid']
        self._threads = []
        self._events_queue = None

        if not self._local_jemu:
            self._events_queue = Queue.Queue()
            self._events_handler_should_run = True
            self._events_handler_thread = threading.Thread(target=self._event_sender)
            self._events_handler_thread.setDaemon(True)
            self._threads.append(self._events_handler_thread)
            self._events_handler_thread.start()

    def upload_file(self, file_path):
        t = threading.Thread(target=self._upload_file, args=[file_path])
        t.start()
        self._threads.append(t)

    def _upload_file(self, filepath):
        filename = os.path.basename(filepath)
        signs = {"$", "#", "[", "]"}
        num_of_signs = filename.count('.') - 1
        filename = filename.replace('.', '_', num_of_signs)
        for i in signs:
            filename = filename.replace(i, '_')

        filename = str(int(calendar.timegm(time.gmtime()))) + '_' + filename
        if self._user_uid is None:
            raise UnInitializedError

        headers = self._headers
        headers['Content-Type'] = 'application/octet-stream'
        res = requests.post(
            '{}/firmwares/upload/{}/{}'.format(self._api_url, self._user_uid, filename),
            headers=headers
        )
        res.raise_for_status()
        signed_url = res.text
        with open(filepath, 'rb') as data:
            requests.put(signed_url, data=data)

    # def check_status(self, filename):
    #     if self._user_uid is None:
    #         raise UnInitializedError
    #
    #     headers = self._headers
    #     headers['Content-Type'] = 'application/text'
    #     res = requests.get(
    #         '{}/firmwares/{}/{}/status'.format(self._api_url, self._user_uid, filename),
    #         headers=headers
    #     )
    #     return res

    def send_event(self, event_dict):
        if self._user_uid is None:
            raise UnInitializedError
        event_name = event_dict['event']
        event_labels = {}
        if "labels" in event_dict:
            event_labels = event_dict["labels"]
        event_labels['distinct_id'] = self._user_uid
        event_labels['sdk_version'] = jumper_current_version
        event_labels['Operating System'] = sys.platform
        headers = self._headers
        headers['Content-Type'] = 'application/json'
        res = requests.post(
            '{}/analytics/{}/{}'.format(self._api_url, self._user_uid, event_name),
            headers=headers,
            json=event_labels
        )
        return res

    # def check_error(self, filename):
    #     if self._user_uid is None:
    #         raise UnInitializedError
    #
    #     headers = self._headers
    #     headers['Content-Type'] = 'application/text'
    #     res = requests.get(
    #         '{}/firmwares/{}/{}/error'.format(self._api_url, self._user_uid, filename),
    #         headers=headers
    #     )
    #     return res

    def add_event(self, message):
        if self._events_queue:
            self._events_queue.put(message)

    def _event_sender(self):
        while True:
            try:
                cur_event = self._events_queue.get(True, timeout=0.2)
            except Queue.Empty:
                if not self._events_handler_should_run:
                    return
                else:
                    continue
            retries = 3
            while retries > 0:
                try:
                    self.send_event(cur_event)
                except Exception:
                    retries -= 1
                    sleep(1)
                else:
                    break

    def stop(self):
        self._events_handler_should_run = False
        self._stop_threads()

    def _stop_threads(self):
        for t in self._threads:
            if t.is_alive():
                t.join()



import sys, os

CORE_LINUX_OS = "linux"
CORE_MAC_OS = "darwin"

JEMU_LINUX_DIR = "jemu-linux"
JEMU_MAC_DIR = "jemu-mac"
JEMU_WINDOWS_DIR = "jemu-windows"

HERE = os.path.abspath(os.path.dirname(__file__))

JEMU_DIR = None
if sys.platform.startswith(CORE_LINUX_OS):
    JEMU_DIR = os.path.join(HERE, 'jemu', JEMU_LINUX_DIR)
elif sys.platform.startswith(CORE_MAC_OS):
    JEMU_DIR = os.path.join(HERE, 'jemu', JEMU_MAC_DIR)
elif os.name == 'nt':
    JEMU_DIR = os.path.join(HERE, 'jemu', JEMU_WINDOWS_DIR)


def _get_file_path(filename_without_extension):
    filename = filename_without_extension
    if os.name == 'nt':
        filename += '.exe'
    return os.path.join(JEMU_DIR, filename)


JEMU_PATH = _get_file_path('jemu')
OBJCOPY_PATH = _get_file_path('arm-none-eabi-objcopy')
OBJDUMP_PATH = _get_file_path('arm-none-eabi-objdump')



import tornado.ioloop
import tornado.web
import json
import webbrowser
import socket
import signal
import sys

from .config import set_token, DEFAULT_CONFIG

URL = "https://vlab.jumper.io"
# URL = "http://localhost:3000"
CLI_LOGIN_URL = URL + '/cli_login'
GET_STARTED_URL = URL + '/getstarted'
token = None

close_tornado = False


class TokenHandler(tornado.web.RequestHandler):
    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "*")
        self.set_header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')

    def get(self):
        global token
        token = self.request.query_arguments['code'][0]
        self.write('Login Success, please close the tab/windows.')
        self.set_status(200)
        tornado.ioloop.IOLoop.current().stop()

    def post(self):
        global token
        data = json.loads(self.request.body)
        token = data['token']
        self.write('OK')
        tornado.ioloop.IOLoop.current().stop()

    def options(self):
        self.set_status(204)
        self.finish()


def try_exit():
    if close_tornado:
        tornado.ioloop.IOLoop.current().stop()


def sig_handler(sig, frame):
    global close_tornado
    close_tornado = True


def token_server():
    global close_tornado
    global token

    token = None
    app = tornado.web.Application([(r"/token", TokenHandler)])
    port = choose_port()
    server = app.listen(port)
    webbrowser.get()
    webbrowser.open_new('{}/{}'.format(CLI_LOGIN_URL, port))

    tornado.ioloop.PeriodicCallback(try_exit, 100).start()

    close_tornado = False
    signal.signal(signal.SIGTERM, sig_handler)
    signal.signal(signal.SIGINT, sig_handler)
    tornado.ioloop.IOLoop.current().start()

    server.stop()

    if token is None:
        raise ValueError('Could not get token on login')
    return token


def choose_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('localhost', 0))
    addr, port = s.getsockname()
    s.close()
    return port


def login(no_browser=False):
    global token
    try:
        if no_browser:
            raise webbrowser.Error
        token = token_server()
    except webbrowser.Error:
        print("Could not open browser for automatic login, fallback to manual login.")
        print("Please visit {} to get your token.".format(GET_STARTED_URL))

        from builtins import input
        new_token = input("Enter token: ")
        if new_token == '':
            raise ValueError("No token was received")
        token = new_token

    set_token(token)
    print("Token has been saved to {}".format(DEFAULT_CONFIG))
    print("Login successful")
    return token


if __name__ == "__main__":
    print(token_server())
